#include <Arduino.h>
#include <stdio.h>
#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_ILI9341.h>
#include <pgmspace.h>
#include "picojpeg.h"

uint8_t number = 0;

#define SPI_miso    36
#define SPI_mosi    35
#define SPI_sck     37
#define SPI_sd_ss    8
#define SPI_tft_ss  38
#define SPI_tft_dc  14
#define SPI_tft_rst  9

Adafruit_ILI9341 tft = Adafruit_ILI9341(SPI_tft_ss, SPI_tft_dc, SPI_mosi, SPI_sck, SPI_tft_rst, SPI_miso);

extern const unsigned char d0[], d1[], d2[], d3[], d4[], d5[], d6[], d7[], d8[], d9[];
extern const unsigned long d0len, d1len, d2len, d3len, d4len, d5len, d6len, d7len, d8len, d9len;

#define jpgColor(c)                                  \
  (((uint16_t)(((uint8_t *)(c))[0] & 0xF8) << 8) |   \
   ((uint16_t)(((uint8_t *)(c))[1] & 0xFC) << 3) |   \
   ((((uint8_t *)(c))[2] & 0xF8) >> 3))

#define local_min(a,b)  (((a) < (b)) ? (a) : (b))

int digit_width  = 146;
int digit_height = 240;
uint8_t *digits[10];

// for JPEGDec - private
FILE *fileJpeg;
pjpeg_image_info_t image_info;
int jpeg_is_available;
int jpeg_mcu_x;
int jpeg_mcu_y;
uint jpeg_read_offset;
uint jpeg_size;

// for JPEGDec - public
uint8_t *jpeg_pImage;
int jpeg_width;
int jpeg_height;
int jpeg_MCUSPerRow;
int jpeg_MCUSPerCol;
int jpeg_MCUWidth;
int jpeg_MCUHeight;
int jpeg_MCUx;
int jpeg_MCUy;

static uint8_t pjpeg_need_bytes_callback(uint8_t* pBuf, uint8_t buf_size, uint8_t *pBytes_actually_read, void *pCallback_data) {
	pjpeg_callback(pBuf, buf_size, pBytes_actually_read, pCallback_data);
	return 0;
}

uint8_t pjpeg_callback(uint8_t* pBuf, uint8_t buf_size, uint8_t *pBytes_actually_read, void *pCallback_data) {
	uint n;

	n = local_min(jpeg_size - jpeg_read_offset, buf_size);

	fread(pBuf,1,n,fileJpeg);

	*pBytes_actually_read = (uint8_t)(n);
	jpeg_read_offset += n;
	return 0;
}


int JpegDec_decode(unsigned long data_size){

	jpeg_mcu_x = 0;
	jpeg_mcu_y = 0;
	jpeg_is_available = 0;

	jpeg_read_offset = 0;

	jpeg_size = data_size;

	uint8_t status = pjpeg_decode_init(&image_info, pjpeg_need_bytes_callback, NULL, 0);
            
	if (status) {
		Serial.print("pjpeg_decode_init() failed");
		return -1;
	}

	// In reduce mode output 1 pixel per 8x8 block.
	jpeg_width = image_info.m_width;
	jpeg_height = image_info.m_height;
	jpeg_MCUSPerRow = image_info.m_MCUSPerRow;
	jpeg_MCUSPerCol = image_info.m_MCUSPerCol;
	jpeg_MCUWidth   = image_info.m_MCUWidth;
	jpeg_MCUHeight  = image_info.m_MCUHeight;

	// Image Information
	Serial.print("jpeg_width :");
	Serial.println(jpeg_width);
	Serial.print("jpeg_height :");
	Serial.println(jpeg_height);
	Serial.print("jpeg_MCUSPerRow :");
	Serial.println(jpeg_MCUSPerRow);
	Serial.print("jpeg_MCUSPerCol :");
	Serial.println(jpeg_MCUSPerCol);
	Serial.print("jpeg_MCUWidth :");
	Serial.println(jpeg_MCUWidth);
	Serial.print("jpeg_MCUHeight :");
	Serial.println(jpeg_MCUHeight);
	Serial.println("");

	//pImage = (uint8 *)malloc(image_info.m_MCUWidth * image_info.m_MCUHeight * 3s);
	jpeg_pImage = new uint8_t[image_info.m_MCUWidth * image_info.m_MCUHeight * 3];
	if (!jpeg_pImage) {
		fclose(fileJpeg);
		Serial.println("Memory Allocation Failure");
		return -1;
	}
	memset(jpeg_pImage , 0 , sizeof(jpeg_pImage));

	jpeg_is_available = 1 ;

	return JpegDec_decode_mcu();
}

int JpegDec_decode_mcu(void) {

	uint8_t status = pjpeg_decode_mcu();

	if (status) {
		jpeg_is_available = 0 ;

		if (status != PJPG_NO_MORE_BLOCKS) {
			Serial.print("pjpeg_decode_mcu() failed with status ");
			return -1;
		}
	}
	return 1;
}


int JpegDec_read(void) {
	int y, x;
	uint8_t *pDst_row;

	if(jpeg_is_available == 0 || jpeg_mcu_y >= image_info.m_MCUSPerCol) return 0;

	// Copy MCU's pixel blocks into the destination pixels.
	pDst_row = jpeg_pImage;
	for (y = 0; y < image_info.m_MCUHeight; y += 8) {
		const int by_limit = local_min(8, image_info.m_height - (jpeg_mcu_y * image_info.m_MCUHeight + y));

		for (x = 0; x < image_info.m_MCUWidth; x += 8) {
			uint8_t *pDst_block = pDst_row + x * 3;

			// Compute source byte offset of the block in the decoder's MCU buffer.
			uint src_ofs = (x * 8U) + (y * 16U);
			const uint8_t *pSrcR = image_info.m_pMCUBufR + src_ofs;
			const uint8_t *pSrcG = image_info.m_pMCUBufG + src_ofs;
			const uint8_t *pSrcB = image_info.m_pMCUBufB + src_ofs;

			const int bx_limit = local_min(8, image_info.m_width - (jpeg_mcu_x * image_info.m_MCUWidth + x));

			int bx, by;
			for (by = 0; by < by_limit; by++) {
				uint8_t *pDst = pDst_block;

				for (bx = 0; bx < bx_limit; bx++) {
					pDst[0] = *pSrcR++;
					pDst[1] = *pSrcG++;
					pDst[2] = *pSrcB++;
					pDst += 3;
				}

				pSrcR += (8 - bx_limit);
				pSrcG += (8 - bx_limit);
				pSrcB += (8 - bx_limit);
				pDst_block += jpeg_MCUWidth * 3;
			}
		}
		pDst_row += (jpeg_MCUWidth * 8 *3 );
	}

	jpeg_MCUx = jpeg_mcu_x;
	jpeg_MCUy = jpeg_mcu_y;
    
	jpeg_mcu_x++;
	if (jpeg_mcu_x == image_info.m_MCUSPerRow) {
		jpeg_mcu_x = 0;
		jpeg_mcu_y++;
	}

	if(JpegDec_decode_mcu() < 0) jpeg_is_available = 0 ;

	return 1;
}

void setup() {

	int key = -1;
  Serial.begin(115200);

  tft.begin();
  tft.setRotation(3);
  tft.fillScreen(ILI9341_BLACK);

	for (int digit=0;digit<10;digit++) {
		digits[digit] = (uint8_t *)malloc(digit_width * digit_height * 3);
		if (!digits[digit]) {
			Serial.println("digits PSRAM Allocation Failure");
			return;
		}
		switch (digit) {
			case 0: Jpg2Pixel(digit, d0, d0len); break;
			case 1: Jpg2Pixel(digit, d1, d1len); break;
			case 2: Jpg2Pixel(digit, d2, d2len); break;
			case 3: Jpg2Pixel(digit, d3, d3len); break;
			case 4: Jpg2Pixel(digit, d4, d4len); break;
			case 5: Jpg2Pixel(digit, d5, d5len); break;
			case 6: Jpg2Pixel(digit, d6, d6len); break;
			case 7: Jpg2Pixel(digit, d7, d7len); break;
			case 8: Jpg2Pixel(digit, d8, d8len); break;
			case 9: Jpg2Pixel(digit, d9, d9len); break;
		}
	}
	while (true) {
		if (number>99) number = 0;
		int d1 = number / 10;
		int d2 = number % 10;
		if (key!=d1) {
			tft.fillScreen(ILI9341_BLACK);
			showDigit(d1, 14, 0);
			key = d1;
		}
		showDigit(d2, 160, 0);
//	delay(1000);
		number++;
	}
}

void loop() {
}

boolean Jpg2Pixel(int digit, const unsigned char *data, const unsigned long len) {

	int pos;
	uint8_t *pImg;
	int x,y,bx,by;
	uint8_t *pixels;

	pixels = digits[digit];

	if((fileJpeg = fmemopen((void*)data, len, "rb"))==NULL) {
		Serial.print("fmemopen() failed");
		return false;
	} else {
		Serial.println("fmemopen()");
	}

	// Decoding start
	if (JpegDec_decode(len) < 0) return false;

	jpeg_pImage = new uint8_t[jpeg_MCUWidth * jpeg_MCUHeight * 3];
	if (!jpeg_pImage) {
		fclose(fileJpeg);
		Serial.println("Memory Allocation Failure");
		return false;
	}
	memset(jpeg_pImage, 0, sizeof(jpeg_pImage));

	while(JpegDec_read()){
		pImg = jpeg_pImage;

		for(by=0; by<jpeg_MCUHeight; by++){
			for(bx=0; bx<jpeg_MCUWidth; bx++){
				x = jpeg_MCUx * jpeg_MCUWidth + bx;
				y = jpeg_MCUy * jpeg_MCUHeight + by;
				if(x<jpeg_width && y<jpeg_height){
					pos = (x + (y * jpeg_width)) * 3;
					pixels[pos  ] = pImg[0];
					pixels[pos+1] = pImg[1];
					pixels[pos+2] = pImg[2];
				}
				pImg += 3;
			}
		}
	}
	fclose(fileJpeg);
	delete[] jpeg_pImage;

	return true;
}

void showDigit(int digit, int offset_x, int offset_y) {

	uint8_t *pImg;
	int x,y;

	pImg = digits[digit];
	for(y=0; y<digit_height; y++){
		for(x=0; x<digit_width; x++){
			tft.drawPixel(x+offset_x, y+offset_y, tft.color565(pImg[0], pImg[1], pImg[2]));
			pImg += 3;
		}
	}
}
